﻿
Imports System.Net
Imports System.Net.WebRequestMethods
Imports System.IO
Imports System.Text.RegularExpressions


Module Utils_FTP

    Friend Working As Boolean = False
    Friend AbortFlag As Boolean
    Friend Delegate Sub ProgressCallBack(ByVal params As ProgressParams)

    Friend Structure ProgressParams
        Dim progress As Double
        Dim fileLength As Long
        Dim bytesRead As Int32
        Dim bytesTransferred As Long
        Dim secondsTotal As Double
        Dim secondsElapsed As Double
        Dim secondsRemaining As Double
        Dim kbps As Double
        Dim timeStart As Date
        Dim elapsedTime As Double
        Dim totalTime As Double
    End Structure

    Friend Structure FileParams
        Dim CompletePath As String
        Dim FileName As String
        Dim Length As Long
        Dim Type As String
        Dim Modified As String
        Dim users As String
        Dim permissions As String
    End Structure




    ' ===================================================================================================
    '   FTP_ListDirectoryDetails
    ' ===================================================================================================

    Friend Function FTP_ListDirectoryDetails(ByVal ServerDirectory As String, _
                                             ByVal UserName As String, _
                                             ByVal Password As String) As List(Of FileParams)

        Dim items As List(Of FileParams)
        items = New List(Of FileParams)

        If working Then Return items
        working = True
        Application.UseWaitCursor = True
        Cursor.Current = Cursors.WaitCursor

        Dim FtpRequest As FtpWebRequest = Nothing
        Dim FtpResponse As FtpWebResponse = Nothing
        Dim FtpStream As Stream = Nothing
        Dim reader As StreamReader = Nothing

        If Not ServerDirectory.EndsWith("/") Then ServerDirectory &= "/"

        Try
            ' --------------------------------------------------------------------- FPT request (ListDirectory)
            FtpRequest = TryCast(WebRequest.Create(ServerDirectory), FtpWebRequest)
            FtpRequest.Credentials = New NetworkCredential(UserName, Password)
            FtpRequest.Method = WebRequestMethods.Ftp.ListDirectoryDetails
            FtpRequest.UsePassive = True
            FtpRequest.UseBinary = True
            FtpRequest.KeepAlive = False ' <<< close the connection when done

            ' --------------------------------------------------------------------- get the FTP stream
            FtpResponse = TryCast(FtpRequest.GetResponse, FtpWebResponse)
            FtpStream = FtpResponse.GetResponseStream()

            ' --------------------------------------------------------------------- get the stream reader
            reader = New StreamReader(FtpStream)

            ' --------------------------------------------------------------------- start the read loop
            Dim line As String
            Dim fp As FileParams = Nothing
            While Not reader.EndOfStream
                ' ----------------------------------------------------------------- read line
                line = reader.ReadLine()
                'Debug.Print(line)
                ' ----------------------------------------------------------------- decode line
                fp = DecodeDirectoryLine_UsingRegularExpressions(line)
                ' ----------------------------------------------------------------- change invalid characters
                fp.FileName = Replace(fp.FileName, "\", "%5C")
                fp.FileName = Replace(fp.FileName, "�", "à")
                ' ----------------------------------------------------------------- add item
                If fp.FileName IsNot Nothing Then
                    If Not fp.FileName.StartsWith(".") Then
                        fp.CompletePath = ServerDirectory & fp.FileName
                        items.Add(fp)
                    End If
                End If
            End While

        Catch ex As Exception
            MsgBox(ex.Message)

        Finally
            If FtpStream IsNot Nothing Then FtpStream.Close()
            If FtpResponse IsNot Nothing Then FtpResponse.Close()
            If reader IsNot Nothing Then reader.Close()

        End Try

        Application.UseWaitCursor = False
        Cursor.Current = Cursors.Default
        Working = False
        Return items
    End Function



    ' ===================================================================================================
    '   DecodeDirectoryLine_UsingRegularExpressions
    ' ===================================================================================================

    Private Function DecodeDirectoryLine_UsingRegularExpressions(ByVal line As String) As FileParams
        Dim fp As FileParams
        fp = New FileParams
        ' ------------------------------------------------------------------ parse line
        Dim m As Match = GetMatchingRegex(line)
        If m Is Nothing Then
            ' -------------------------------------------------------------- failed
            MsgBox("Unable to parse line:    """ & line & """" & vbCr & vbCr & _
                   "Please edit the file ""RegExp.txt", MsgBoxStyle.Information)
        Else
            fp.FileName = m.Groups("name").Value
            fp.Length = CLng(Val(m.Groups("size").Value))
            fp.permissions = m.Groups("permission").Value
            fp.users = m.Groups("usernum").Value '& m.Groups("users").Value
            ' -------------------------------------------------------------- it is a Folder ?
            Dim _dir As String = m.Groups("dir").Value
            If (_dir <> "" And _dir <> "-") Then
                fp.Type = "dir"
            Else
                fp.Type = "file"
            End If
            ' -------------------------------------------------------------- decode timestamp
            fp.Modified = m.Groups("timestamp").Value
            fp.Modified = RemoveMultipleSpaces(fp.Modified)
            fp.Modified = DecodeDateAndTime(fp.Modified)
        End If
        Return fp
    End Function

    Private Function GetMatchingRegex(ByVal line As String) As Match
        Dim rx As Regex
        Dim m As Match
        For i As Integer = 0 To RegExp.Length - 1
            rx = New Regex(RegExp(i))
            m = rx.Match(line)
            If m.Success Then
                'Debug.Print(i.ToString & vbTab & line)
                'Debug.Print(line)
                Return m
            End If
        Next
        Return Nothing
    End Function

   
    Private Function RemoveMultipleSpaces(ByVal s As String) As String
        While s.Contains("  ")
            s = s.Replace("  ", " ")
        End While
        Return s
    End Function

    Private Function DecodeDateAndTime(ByVal dt As String) As String
        Dim d As Date
        If dt.Contains(":") Then
            Try
                d = Date.ParseExact(dt, "MMM d HH:mm", Globalization.CultureInfo.InvariantCulture)
            Catch
                '    "08-13-15 11:04AM"
                'dt = "08-13-15 11:04PM"
                d = Date.ParseExact(dt, "MM-dd-yy hh:mmtt", Globalization.CultureInfo.InvariantCulture)
            End Try
            If d > Date.Now Then d = d.AddYears(-1)
            ' -------------------------------------- convert to string using generic format "dd/MM/yyyy HH:mm"
            Return d.ToString("g")
        Else
            d = Date.ParseExact(dt, "MMM d yyyy", Globalization.CultureInfo.InvariantCulture)
            ' -------------------------------------- convert to string using generic format "dd/MM/yyyy"
            Return d.ToString("d")
        End If
    End Function




    ' ===================================================================================================
    '   FTP_Upload
    ' ===================================================================================================

    Friend Sub FTP_Upload(ByVal LocalFile As String, _
                           ByVal ServerFile As String, _
                           ByVal UserName As String, _
                           ByVal Password As String, _
                           Optional ByVal Callback As ProgressCallBack = Nothing)

        If working Then Exit Sub
        working = True
        Application.UseWaitCursor = True
        Cursor.Current = Cursors.WaitCursor

        Dim params As ProgressParams
        Dim FileStream As FileStream = Nothing
        Dim FtpStream As Stream = Nothing
        Dim FtpRequest As FtpWebRequest = Nothing

        With params

            Try
                ' --------------------------------------------------------------------- init buffer
                Dim bytesPerBlock As Integer = 4096
                Dim buffer(bytesPerBlock - 1) As Byte

                ' --------------------------------------------------------------------- FPT request (upload)
                FtpRequest = TryCast(WebRequest.Create(ServerFile), FtpWebRequest)
                FtpRequest.Method = WebRequestMethods.Ftp.UploadFile
                FtpRequest.Credentials = New NetworkCredential(UserName, Password)
                FtpRequest.UsePassive = True
                FtpRequest.UseBinary = True
                FtpRequest.KeepAlive = False ' <<< close the connection when done

                ' --------------------------------------------------------------------- open the FTP stream
                FtpStream = FtpRequest.GetRequestStream()

                ' --------------------------------------------------------------------- open the file stream
                FileStream = New FileStream(LocalFile, _
                                            FileMode.Open, _
                                            FileAccess.Read)

                ' --------------------------------------------------------------------- start the write loop
                AbortFlag = False
                .fileLength = FileStream.Length
                .bytesTransferred = 0
                .timeStart = Date.Now
                Dim oldseconds As Double = 0
                Do
                    ' ----------------------------------------------------------------- read from the file
                    .bytesRead = FileStream.Read(buffer, 0, buffer.Length)
                    '
                    If .bytesRead > 0 Then

                        ' ------------------------------------------------------------- update total bytes
                        .bytesTransferred += .bytesRead

                        ' ------------------------------------------------------------- write to stream
                        FtpStream.Write(buffer, 0, .bytesRead)

                        ' ------------------------------------------------------------- update progress
                        If Callback IsNot Nothing Then
                            .progress = (.bytesTransferred * 100) \ .fileLength
                            .secondsElapsed = (Date.Now - .timeStart).TotalSeconds
                            .secondsTotal += 0.3 * (.secondsElapsed * .fileLength / .bytesTransferred - .secondsTotal)
                            .secondsRemaining = .secondsTotal - .secondsElapsed
                            If .secondsTotal > 0 Then .kbps = 8 / 1000 * .fileLength / .secondsTotal
                            If .secondsElapsed >= oldseconds + 0.1 Then
                                oldseconds = .secondsElapsed
                                Callback(params)
                            End If
                        End If
                    End If

                    If AbortFlag Then FtpRequest.Abort()

                    Application.DoEvents()

                Loop While .bytesRead > 0


            Catch ex As Exception
                MsgBox(ex.Message)

            Finally
                If Callback IsNot Nothing Then
                    .progress = 0
                    Callback(params)
                End If
                If FileStream IsNot Nothing Then FileStream.Close()
                If FtpStream IsNot Nothing Then FtpStream.Close()

            End Try

        End With

        Application.UseWaitCursor = False
        Cursor.Current = Cursors.Default
        working = False
    End Sub




    ' ===================================================================================================
    '   FTP_Download
    ' ===================================================================================================

    Friend Sub FTP_Download(ByVal LocalFile As String, _
                            ByVal ServerFile As String, _
                            ByVal UserName As String, _
                            ByVal Password As String, _
                            Optional ByVal Length As Long = -1, _
                            Optional ByVal Callback As ProgressCallBack = Nothing)

        If Working Then Exit Sub
        Working = True
        Application.UseWaitCursor = True
        Cursor.Current = Cursors.WaitCursor

        Dim params As ProgressParams
        Dim FtpRequest As FtpWebRequest = Nothing
        Dim FtpResponse As FtpWebResponse = Nothing
        Dim FtpStream As Stream = Nothing
        Dim LocaleFileStream As FileStream = Nothing

        With params

            Try
                ' --------------------------------------------------------------------- init buffer
                Dim bytesPerBlock As Integer = 4096
                Dim buffer(bytesPerBlock - 1) As Byte

                ' --------------------------------------------------------------------- File Length via FPT request
                If Length < 0 Then
                    ' ----------------------------------------------------------------- FPT request (get file size)
                    FtpRequest = TryCast(WebRequest.Create(ServerFile), FtpWebRequest)
                    FtpRequest.Credentials = New NetworkCredential(UserName, Password)
                    FtpRequest.Method = WebRequestMethods.Ftp.GetFileSize
                    FtpRequest.UsePassive = True
                    FtpRequest.UseBinary = True
                    FtpRequest.KeepAlive = True ' <<< don't close the connection
                    ' ----------------------------------------------------------------- try to get the file length
                    Try
                        FtpResponse = TryCast(FtpRequest.GetResponse, FtpWebResponse)
                        .fileLength = FtpResponse.ContentLength
                    Catch ex As Exception
                        .fileLength = 0
                    End Try
                Else
                    .fileLength = Length
                End If

                ' --------------------------------------------------------------------- FPT request (download)
                FtpRequest = TryCast(WebRequest.Create(ServerFile), FtpWebRequest)
                FtpRequest.Credentials = New NetworkCredential(UserName, Password)
                FtpRequest.Method = WebRequestMethods.Ftp.DownloadFile
                FtpRequest.UsePassive = True
                FtpRequest.UseBinary = True
                FtpRequest.KeepAlive = False ' <<< close the connection when done
                ' --------------------------------------------------------------------- get the FTP stream
                FtpResponse = TryCast(FtpRequest.GetResponse, FtpWebResponse)
                FtpStream = FtpResponse.GetResponseStream()

                ' --------------------------------------------------------------------- open the locale file stream
                LocaleFileStream = New FileStream(LocalFile, _
                                                  FileMode.OpenOrCreate, _
                                                  FileAccess.Write)

                ' --------------------------------------------------------------------- start the read loop
                AbortFlag = False
                .bytesTransferred = 0
                .timeStart = Date.Now
                Dim oldseconds As Double = 0
                Do
                    ' ----------------------------------------------------------------- read from ftp
                    .bytesRead = FtpStream.Read(buffer, 0, buffer.Length)

                    If .bytesRead > 0 Then

                        ' ------------------------------------------------------------- update total bytes
                        .bytesTransferred += .bytesRead

                        ' ------------------------------------------------------------- write to local file
                        LocaleFileStream.Write(buffer, 0, .bytesRead)

                        ' ------------------------------------------------------------- update progress
                        If Callback IsNot Nothing Then
                            If .fileLength > 0 Then .progress = (.bytesTransferred * 100) \ .fileLength
                            .secondsElapsed = (Date.Now - .timeStart).TotalSeconds
                            .secondsTotal += 0.3 * (.secondsElapsed * .fileLength / .bytesTransferred - .secondsTotal)
                            .secondsRemaining = .secondsTotal - .secondsElapsed
                            If .secondsTotal > 0 Then .kbps = 8 / 1000 * .fileLength / .secondsTotal
                            If .secondsElapsed >= oldseconds + 0.1 Then
                                oldseconds = .secondsElapsed
                                Callback(params)
                            End If
                        End If
                    End If

                    If AbortFlag Then FtpRequest.Abort()

                    Application.DoEvents()

                Loop While .bytesRead > 0


            Catch ex As Exception
                MsgBox(ex.Message)

            Finally
                If Callback IsNot Nothing Then
                    .progress = 0
                    Callback(params)
                End If
                If LocaleFileStream IsNot Nothing Then LocaleFileStream.Close()
                If FtpStream IsNot Nothing Then FtpStream.Close()
                If FtpResponse IsNot Nothing Then FtpResponse.Close()

            End Try

        End With

        Application.UseWaitCursor = False
        Cursor.Current = Cursors.Default
        Working = False
    End Sub




    ' ===================================================================================================
    '   HTTP_Download
    ' ===================================================================================================

    Friend Sub HTTP_Download(ByVal LocalFile As String, _
                             ByVal ServerFile As String, _
                             Optional ByVal Callback As ProgressCallBack = Nothing)

        If working Then Exit Sub
        working = True
        Application.UseWaitCursor = True
        Cursor.Current = Cursors.WaitCursor

        Dim params As ProgressParams
        Dim WebRequest As WebRequest = Nothing
        Dim WebResponse As WebResponse = Nothing
        Dim HttpStream As Stream = Nothing
        Dim LocaleFileStream As FileStream = Nothing

        With params

            Try
                ' --------------------------------------------------------------------- init buffer
                Dim bytesPerBlock As Integer = 4096
                Dim buffer(bytesPerBlock - 1) As Byte

                ' --------------------------------------------------------------------- WEB request
                WebRequest = WebRequest.Create(ServerFile)

                ' --------------------------------------------------------------------- Set security protocol to Tls12
                ServicePointManager.SecurityProtocol = DirectCast(3072, SecurityProtocolType)

                ' --------------------------------------------------------------------- WEB response
                WebResponse = WebRequest.GetResponse

                ' --------------------------------------------------------------------- get the file length
                Try
                    .fileLength = WebResponse.ContentLength
                Catch ex As Exception
                    .fileLength = 0
                End Try

                ' --------------------------------------------------------------------- get the HTTP stream
                HttpStream = WebResponse.GetResponseStream()

                ' --------------------------------------------------------------------- open the locale file stream
                LocaleFileStream = New FileStream(LocalFile, _
                                                  FileMode.OpenOrCreate, _
                                                  FileAccess.Write)

                ' --------------------------------------------------------------------- start the read loop
                AbortFlag = False
                .bytesTransferred = 0
                .timeStart = Date.Now
                Dim oldseconds As Double = 0
                Do
                    ' ----------------------------------------------------------------- read from ftp
                    .bytesRead = HttpStream.Read(buffer, 0, buffer.Length)

                    If .bytesRead > 0 Then

                        ' ------------------------------------------------------------- update total bytes
                        .bytesTransferred += .bytesRead

                        ' ------------------------------------------------------------- write to local file
                        LocaleFileStream.Write(buffer, 0, .bytesRead)

                        ' ------------------------------------------------------------- update progress
                        If Callback IsNot Nothing Then
                            If .fileLength > 0 Then .progress = (.bytesTransferred * 100) \ .fileLength
                            .secondsElapsed = (Date.Now - .timeStart).TotalSeconds
                            .secondsTotal += 0.3 * (.secondsElapsed * .fileLength / .bytesTransferred - .secondsTotal)
                            .secondsRemaining = .secondsTotal - .secondsElapsed
                            If .secondsTotal > 0 Then .kbps = 8 / 1000 * .fileLength / .secondsTotal
                            If .secondsElapsed >= oldseconds + 0.1 Then
                                oldseconds = .secondsElapsed
                                Callback(params)
                            End If
                        End If
                    End If

                    If AbortFlag Then WebRequest.Abort()

                    Application.DoEvents()

                Loop While .bytesRead > 0


            Catch ex As Exception
                MsgBox(ex.Message)

            Finally
                If Callback IsNot Nothing Then
                    .progress = 0
                    Callback(params)
                End If
                If LocaleFileStream IsNot Nothing Then LocaleFileStream.Close()
                If HttpStream IsNot Nothing Then HttpStream.Close()
                If WebResponse IsNot Nothing Then WebResponse.Close()

            End Try

        End With

        Application.UseWaitCursor = False
        Cursor.Current = Cursors.Default
        working = False
    End Sub



    ' ===================================================================================================
    '   FTP_DeleteFile
    ' ===================================================================================================

    Friend Sub FTP_DeleteFile(ByVal FileName_WithPath As String, _
                              ByVal UserName As String, _
                              ByVal Password As String)

        If working Then Exit Sub
        working = True
        Application.UseWaitCursor = True
        Cursor.Current = Cursors.WaitCursor

        Dim FtpRequest As FtpWebRequest = Nothing
        Try
            ' --------------------------------------------------------------------- FPT request (ListDirectory)
            FtpRequest = TryCast(WebRequest.Create(FileName_WithPath), FtpWebRequest)
            FtpRequest.Credentials = New NetworkCredential(UserName, Password)
            FtpRequest.UsePassive = True
            FtpRequest.UseBinary = True
            FtpRequest.KeepAlive = False ' <<< close the connection when done
            FtpRequest.Method = WebRequestMethods.Ftp.DeleteFile

            ' --------------------------------------------------------------------- exec and close
            FtpRequest.GetResponse().Close()

        Catch ex As Exception
            MsgBox(ex.Message)
        End Try

        Application.UseWaitCursor = False
        Cursor.Current = Cursors.Default
        working = False
    End Sub


    ' ===================================================================================================
    '   FTP_RemoveDirectory
    ' ===================================================================================================

    Friend Sub FTP_RemoveDirectory(ByVal DirectoryPath As String, _
                                   ByVal UserName As String, _
                                   ByVal Password As String)

        If working Then Exit Sub
        working = True
        Application.UseWaitCursor = True
        Cursor.Current = Cursors.WaitCursor

        Dim FtpRequest As FtpWebRequest = Nothing
        Try
            ' --------------------------------------------------------------------- FPT request (ListDirectory)
            FtpRequest = TryCast(WebRequest.Create(DirectoryPath), FtpWebRequest)
            FtpRequest.Credentials = New NetworkCredential(UserName, Password)
            FtpRequest.UsePassive = True
            FtpRequest.UseBinary = True
            FtpRequest.KeepAlive = False ' <<< close the connection when done
            FtpRequest.Method = WebRequestMethods.Ftp.RemoveDirectory

            ' --------------------------------------------------------------------- exec and close
            FtpRequest.GetResponse().Close()

        Catch ex As Exception
            MsgBox(ex.Message)
        End Try

        Application.UseWaitCursor = False
        Cursor.Current = Cursors.Default
        working = False
    End Sub


    ' ===================================================================================================
    '   FTP_MakeDirectory
    ' ===================================================================================================

    Friend Sub FTP_MakeDirectory(ByVal Directory As String, _
                                   ByVal UserName As String, _
                                   ByVal Password As String)

        If working Then Exit Sub
        working = True
        Application.UseWaitCursor = True
        Cursor.Current = Cursors.WaitCursor

        Dim FtpRequest As FtpWebRequest = Nothing
        Try
            ' --------------------------------------------------------------------- FPT request (ListDirectory)
            FtpRequest = TryCast(WebRequest.Create(Directory), FtpWebRequest)
            FtpRequest.Credentials = New NetworkCredential(UserName, Password)
            FtpRequest.UsePassive = True
            FtpRequest.UseBinary = True
            FtpRequest.KeepAlive = False ' <<< close the connection when done
            FtpRequest.Method = WebRequestMethods.Ftp.MakeDirectory

            ' --------------------------------------------------------------------- exec and close
            FtpRequest.GetResponse().Close()

        Catch ex As Exception
            ' --------------------------------- no exceptions for already existent folders
            'MsgBox(ex.Message)
        End Try

        Application.UseWaitCursor = False
        Cursor.Current = Cursors.Default
        working = False
    End Sub



    ' ===================================================================================================
    '   FTP_RenameFile
    ' ===================================================================================================

    Friend Sub FTP_RenameFile(ByVal OldFileName_WithPath As String, _
                              ByVal NewFileName_WithoutPath As String, _
                              ByVal UserName As String, _
                              ByVal Password As String)

        If working Then Exit Sub
        working = True
        Application.UseWaitCursor = True
        Cursor.Current = Cursors.WaitCursor

        Dim FtpRequest As FtpWebRequest = Nothing
        Try
            ' --------------------------------------------------------------------- FPT request (ListDirectory)
            FtpRequest = TryCast(WebRequest.Create(OldFileName_WithPath), FtpWebRequest)
            FtpRequest.Credentials = New NetworkCredential(UserName, Password)
            FtpRequest.UsePassive = True
            FtpRequest.UseBinary = True
            FtpRequest.KeepAlive = False ' <<< close the connection when done
            FtpRequest.Method = WebRequestMethods.Ftp.Rename

            ' --------------------------------------------------------------------- get the FTP stream
            FtpRequest.RenameTo = NewFileName_WithoutPath

            ' --------------------------------------------------------------------- exec and close
            FtpRequest.GetResponse().Close()

        Catch ex As Exception
            MsgBox(ex.Message)
        End Try

        Application.UseWaitCursor = False
        Cursor.Current = Cursors.Default
        working = False
    End Sub



   

End Module
